@@ -3,14 +3,14 @@
 '''Stateless OpenPGP Scaffolding
 
 Author: Daniel Kahn Gillmor
-Date: October 2019
+Date: April 2022
 License: MIT (see below)
 
 This module implements a pythonic framework for `sop`, the Stateless
 OpenPGP command-line interface.  It makes it easier to build out a
 python-based backend.
 
-  https://tools.ietf.org/html/draft-dkg-openpgp-stateless-cli
+  https://datatracker.ietf.org/doc/draft-dkg-openpgp-stateless-cli/
 
 In particular, sop.StatelessOpenPGP presents a generic baseclass for
 building an implementation of `sop`.
@@ -96,7 +96,7 @@ import logging
 from binascii import unhexlify, hexlify
 from datetime import datetime, timezone
 from argparse import ArgumentParser, _SubParsersAction, Namespace
-from typing import List, Optional, Dict, Sequence, MutableMapping, Tuple, BinaryIO
+from typing import List, Optional, Dict, Sequence, MutableMapping, Tuple, BinaryIO, TYPE_CHECKING
 
 from .__version__ import __version__
 
@@ -123,29 +123,59 @@ class SOPArmorLabel(enum.Enum):
     message = enum.auto()
 
 class SOPException(Exception):
-    exit_code = 2
+    exit_code:int = 2
+    mnemonic:Optional[str] = None
 class SOPNoSignature(SOPException):
     exit_code = 3
+    mnemonic = 'NO_SIGNATURE'
 class SOPAsymmetricAlgorithmUnsupported(SOPException):
     exit_code = 13
+    mnemonic = 'UNSUPPORTED_ASYMMETRIC_ALGO'
 class SOPCertificateNotEncryptionCapable(SOPException):
     exit_code = 17
+    mnemonic = 'CERT_CANNOT_ENCRYPT'
 class SOPMissingRequiredArgument(SOPException):
     exit_code = 19
+    mnemonic = 'MISSING_ARG'
 class SOPIncompleteVerificationInstructions(SOPException):
     exit_code = 23
+    mnemonic = 'INCOMPLETE_VERIFICATION'
 class SOPCouldNotDecrypt(SOPException):
     exit_code = 29
+    mnemonic = 'CANNOT_DECRYPT'
 class SOPNonUTF8Password(SOPException):
     exit_code = 31
+    mnemonic = 'PASSWORD_NOT_HUMAN_READABLE'
 class SOPUnsupportedOption(SOPException):
     exit_code = 37
+    mnemonic = 'UNSUPPORTED_OPTION'
 class SOPInvalidDataType(SOPException):
     exit_code = 41
+    mnemonic = 'BAD_DATA'
 class SOPNotUTF8Text(SOPException):
     exit_code = 53
+    mnemonic = 'EXPECTED_TEXT'
+class SOPOutputExists(SOPException):
+    exit_code = 59
+    mnemonic = 'OUTPUT_EXISTS'
+class SOPMissingInput(SOPException):
+    exit_code = 61
+    mnemonic = 'MISSING_INPUT'
+class SOPKeyIsProtected(SOPException):
+    exit_code = 67
+    mnemonic = 'KEY_IS_PROTECTED'
 class SOPUnsupportedSubcommand(SOPException):
     exit_code = 69
+    mnemonic = 'UNSUPPORTED_SUBCOMMAND'
+class SOPUnsupportedSpecialPrefix(SOPException):
+    exit_code = 71
+    mnemonic = 'UNSUPPORTED_SPECIAL_PREFIX'
+class SOPAmbiguousInput(SOPException):
+    exit_code = 73
+    mnemonic = 'AMBIGUOUS_INPUT'
+class SOPKeyCannotSign(SOPException):
+    exit_code = 79
+    mnemonic = 'KEY_CANNOT_SIGN'
 
 
 class SOPSessionKey(object):
@@ -184,7 +214,7 @@ class SOPSigResult(object):
     def __str__(self) -> str:
         # ensure tz=UTC:
         whendt:datetime = datetime.fromtimestamp(self._when.timestamp(), tz=timezone.utc)
-        when:str = whendt.strftime('%Y-%m-%dT%H:%M%SZ')
+        when:str = whendt.strftime('%Y-%m-%dT%H:%M:%SZ')
         # strip all whitespace from fprs:
         signing_fpr:str = self._signing_fpr.translate(str.maketrans('', '', string.whitespace))
         primary_fpr:str = self._primary_fpr.translate(str.maketrans('', '', string.whitespace))
@@ -221,7 +251,7 @@ class StatelessOpenPGP(object):
         self._parser = ArgumentParser(prog=name, description=description)
         self._parser.add_argument('--debug', action='store_true',
                                   help='show debugging data')
-        _cmds:_SubParsersAction = self._parser.add_subparsers(required=True,
+        _cmds:_SubParsersAction[ArgumentParser] = self._parser.add_subparsers(required=True,
                                             metavar='SUBCOMMAND',
                                             dest='subcmd')
         _subs = {}
@@ -230,8 +260,6 @@ class StatelessOpenPGP(object):
 
         def _add_armor_flag(parser:ArgumentParser) -> None:
             g = parser.add_mutually_exclusive_group(required=False)
-            g.add_argument('--armor', dest='armor', action='store_true',
-                           help='generate ASCII-armored output')
             g.add_argument('--no-armor', dest='armor', action='store_false',
                            help='generate binary output')
             parser.set_defaults(armor=True)
@@ -317,9 +345,6 @@ class StatelessOpenPGP(object):
         _armor.add_argument('--label', choices=SOPArmorLabel.__members__,
                             default='auto',
                             help='specify the type of ASCII armoring')
-        _armor.add_argument('--allow-nested', dest='allow_nested', action='store_true',
-                            help='allow nested ASCII armoring')
-        _armor.set_defaults(allow_nested=False)
 
         _subs['armor'] = _armor
 
@@ -384,7 +409,11 @@ class StatelessOpenPGP(object):
             indirectout.write(data)
         indirectout.close()
 
-    def extend_parsers(self, subcommands:_SubParsersAction,
+    if TYPE_CHECKING:
+        subparsertype = _SubParsersAction[ArgumentParser]
+    else:
+        subparsertype = _SubParsersAction
+    def extend_parsers(self, subcommands:subparsertype,
                        subparsers:Dict[str,ArgumentParser]) -> None:
         '''override this function to add options or subcommands
 
@@ -441,7 +470,7 @@ class StatelessOpenPGP(object):
             out = method(sys.stdin.buffer, **subargs)
             sys.stdout.buffer.write(out)
         except SOPException as e:
-            logging.error(f'{type(e).__name__} {e}')
+            logging.error(f'[{e.mnemonic}] {e}')
             exit(e.exit_code)
 
 
@@ -654,12 +683,10 @@ class StatelessOpenPGP(object):
     def _handle_armor(self,
                       inp:io.BufferedReader,
                       label:str,
-                      allow_nested:bool,
                       **kwargs:Namespace) -> bytes:
         return self.armor(inp.read(), label=SOPArmorLabel.__members__[label],
-                          allow_nested=allow_nested,
                           **kwargs)
-    def armor(self, data:bytes, label:SOPArmorLabel, allow_nested:bool,
+    def armor(self, data:bytes, label:SOPArmorLabel,
               **kwargs:Namespace) -> bytes:
         '''Add OpenPGP ASCII Armor
 
